/*
Copyright 2010-2012 Sergey Levin


This file is part of crosti.

Crosti is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Crosti is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with crosti.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "cimageviewer.h"
#include "cviewportwatcher.h"

#include <QWheelEvent>


CImageViewer::CImageViewer(_schemeData* schemed,QWidget *parent):CViewer(parent),
  _image(),
  schemeData(schemed),
  imageView(new CGraphicsView),
  imageScene(new QGraphicsScene)
{
 setAttribute(Qt::WA_DeleteOnClose);

 imageView->setFrameShape(QFrame::NoFrame);
 imageView->setScene(imageScene);

 if((schemeData!=NULL)&&(!schemeData->isNull())){
     setImage(schemeData->image());
     scaleBar->ui.schemeBtn->setVisible(true);
 }

 layout->insertWidget(0,imageView);

 connect(scaleBar->ui.schemeBtn,SIGNAL(clicked()),SLOT(schemeBtnPress()));
 connect(scaleBar, SIGNAL(scaleChanged(qreal)), imageView, SLOT(setScale(qreal)));
}


CImageViewer::~CImageViewer()
{
  delete imageScene;
  delete imageView;
  delete schemeData;
}


void CImageViewer::setImage(const QImage &img)
{
 _image=img;

 imageScene->clear();
 imageScene->setSceneRect(img.rect());
 imageScene->addPixmap(QPixmap::fromImage(_image));
 scaleBar->ui.scaleSB->setValue(scaleBar->scaleDefault());
 //imageView->setScale(scaleBar->scaleDefault());

 imageView->ViewPortWatcher->setSceneRect(imageScene->sceneRect());

 emit imageChanged(_image);
}


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------


CGraphicsView::CGraphicsView( QWidget *parent):QGraphicsView(parent)
{
    ViewPortWatcher = new CViewPortWatcher(reinterpret_cast<QScrollArea*>(this), this);
}



void CGraphicsView::wheelEvent(QWheelEvent *e)
{
  if(e->modifiers().testFlag(Qt::ControlModifier))e->ignore();
  else QGraphicsView::wheelEvent(e);
}



void CGraphicsView::resizeEvent(QResizeEvent * event )
{
    QGraphicsView::resizeEvent(event);

    ViewPortWatcher->viewPortChanged();
}



void CGraphicsView::setScale(qreal scl)
{
    resetTransform();
    scale(scl,scl);

    ViewPortWatcher->setScale(scl);
}

