/*
Copyright 2010-2012 Sergey Levin


This file is part of crosti.

Crosti is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Crosti is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with crosti.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "imagefilter.h"

#include <QtCore/qmath.h>


_imageFilter::_imageFilter(int edge)
{
 setEdge(edge);
}


void _imageFilter::setEdge(int edge)
{
 _edge=edge;
 setWeight();
}


QImage _imageFilter::filter(QImage image, Name name,QProgressDialog *pb)
{
    //QTime t;
    //t.start();

    setWeight(name);

    QRgb *pixels = reinterpret_cast<QRgb*>(image.bits());

    int width = image.width();
    int height = image.height();
    int weightsize = 2 * _edge + 1;

    for(int i = _edge; i < width - _edge; ++i) {
        for(int j = _edge; j< height - _edge; ++j) {
            qreal red = 0, green = 0, blue = 0;

            for (int k = 0; k < _weight.size(); ++k) {
                qreal w = _weight.at(k);
                div_t dk = div(k, weightsize);
                //d.quot = 1 целая часть;
                //d.rem = 2 остаток от деления;
                int posOffsetted = (j - _edge) * width + (i - _edge);//Начиная с этой позиции применяется матрица фильтра
                int posForWeight = posOffsetted + dk.quot + dk.rem * width;//Позиция на которую надо наложить вес w

                red   += qRed(pixels[posForWeight]) * w;
                green += qGreen(pixels[posForWeight]) * w;
                blue  += qBlue(pixels[posForWeight]) * w;
            }

            int pos = j * width + i;
            pixels[pos] = qRgb(qBound(0, qRound(red/_div), 255),
                               qBound(0, qRound(green/_div), 255),
                               qBound(0, qRound(blue/_div), 255));
        }

        if (pb) {
            if (pb->wasCanceled()) return QImage();
            pb->setValue(pb->value() + 1);
        }
    }


/*
 setWeight(mode);

 for(int i=_edge;i<image.width()-_edge;i++){
   for(int j=_edge;j<image.height()-_edge;j++){

     QList<QRgb>pixel;
     for(int ik=-_edge;ik<_edge+1;ik++)
       for(int jk=-_edge;jk<_edge+1;jk++)
         pixel.append(image.pixel(i+ik,j+jk));


     int clR=0,clG=0,clB=0;
     for(int k=0;k<_weight.size();k++){
         clR+=_weight.at(k)*qRed(pixel.at(k));
         clG+=_weight.at(k)*qGreen(pixel.at(k));
         clB+=_weight.at(k)*qBlue(pixel.at(k));
     }


     clR/=_div;
     clG/=_div;
     clB/=_div;

     if(clR>255)clR=255;
     if(clG>255)clG=255;
     if(clB>255)clB=255;
     if(clR<0)clR=0;
     if(clG<0)clG=0;
     if(clB<0)clB=0;

     image.setPixel(i,j,qRgb(clR,clG,clB));

      if(pb!=0)
        if(pb->wasCanceled())return QImage();

    }

   if(pb!=0)pb->setValue(pb->value()+1);

 }
*/

    //qDebug("Time elapsed: %d ms", t.elapsed());

    return image;
}



QImage _imageFilter::smooth(const QImage &image)
{
    //Уменьшаем, а затем увеличиваем изображение до исходного р-ра - это простейший способ удалить шум
    QImage result = image.scaled(image.size() / 2, Qt::KeepAspectRatio, Qt::SmoothTransformation);
    return result.scaled(image.size(), Qt::KeepAspectRatio, Qt::SmoothTransformation);
}



QImage _imageFilter::medium(QImage image,QProgressDialog *pb)
{
    //QTime t;
    //t.start();

/*
  for(int i=0;i<image.width();i++){
    for(int j=0;j<image.height();j++){
      int i1=image.pixelIndex(i,j);
      if(j<image.height()-3){
        int i2=image.pixelIndex(i,j+2);
        if(i1==i2)image.setPixel(i,j+1,i1);
      }
      if(i<image.width()-3){
        int i2=image.pixelIndex(i+2,j);
        if(i1==i2)image.setPixel(i+1,j,i1);
      }

      if(pb!=0)
        if(pb->wasCanceled())return QImage();
    }
    if(pb!=0)pb->setValue(pb->value()+1);
  }
*/


    uchar *pixels = image.bits();
    int width = image.width();
    int height = image.height();

    for(int i = 0; i < width - 3; ++i) {
        for(int j = 0; j< height - 3; ++j) {

            int pos = j * width + i;
            uchar pixel = pixels[pos];

            if (pixel == pixels[(j + 2) * width + i])
                pixels[(j + 1) * width + i] = pixel;

            if (pixel == pixels[j * width + i + 2])
                pixels[j * width + i + 1] = pixel;
        }

        if (pb) {
            if (pb->wasCanceled()) return QImage();
            pb->setValue(pb->value() + 1);
        }
    }

    //qDebug("Time elapsed: %d ms", t.elapsed());

    return image;

}

QImage _imageFilter::median(QImage image/*,QVector<QRgb>colorTable*/, QProgressDialog *pb)
{
    //QTime t;
    //t.start();
/*
  for(int i=0;i<image.width();i++){
    QList<int>index;
    QList<int>mid;
    mid.append(image.pixelIndex(i,0));
    mid.append(image.pixelIndex(i,0));
    for(int j=0;j<image.height();j++){
      if(j<image.height()-1)mid.append(image.pixelIndex(i,j+1));
      else mid.append(image.pixelIndex(i,j));
      if(j!=0)mid.removeFirst();
      index.append(MIDfilter(colorTable,mid));
    }

    for(int j=0;j<image.height();j++)image.setPixel(i,j,index.at(j));
    if(pb!=0){
      pb->setValue(pb->value()+1);
      if(pb->wasCanceled())return QImage();
    }
  }

  for(int i=0;i<image.height();i++){
    QList<int>index;
    QList<int>mid;
    mid.append(image.pixelIndex(0,i));
    mid.append(image.pixelIndex(0,i));
    for(int j=0;j<image.width();j++){
      if(j<image.width()-1)mid.append(image.pixelIndex(j+1,i));
      else mid.append(image.pixelIndex(j,i));
      if(j!=0)mid.removeFirst();
      index.append(MIDfilter(colorTable,mid));
    }

    for(int j=0;j<image.width();j++)image.setPixel(j,i,index.at(j));
    if(pb!=0){
      pb->setValue(pb->value()+1);
      if(pb->wasCanceled())return QImage();
    }
  }
*/

    uchar *pixels = image.bits();
    int width = image.width();
    int height = image.height();


    QList<uchar> mid;
    int sz = 1;
    //int middle = 2 * sz * (sz + 1);
    //int middle = 2 * sz;
    int middle = sz;
    for (int i = 0; i < width * height; ++i) {
        QList<uchar> view;

        //for (int k = -sz; k < sz + 1; ++k)
        //    view.append(pixels[qBound(0, i + k, width * height - 1)]);
        //for (int k = -sz; k < sz + 1; ++k)
        //    if (k) view.append(pixels[qBound(0, i + k * width, width * height - 1)]);

        for (int k = -sz; k < sz + 1; ++k)
            view.append(pixels[qBound(0, i + k * width, width * height - 1)]);

        //for (int k = -sz; k < sz + 1; ++k)
        //    for (int l = -sz; l < sz + 1; ++l)
        //        view.append(pixels[qBound(0, i + k + l * width, width * height - 1)]);



        qSort(view.begin(), view.end());

        mid.append(view.at(middle));

        if (pb && i % height == 0) {
            if (pb->wasCanceled()) return QImage();
            pb->setValue(pb->value() + 1);
        }
    }


    for (int i = 0; i < width * height; ++i)
        pixels[i] = mid.at(i);


    //qDebug("Time elapsed: %d ms", t.elapsed());

    return image;
}

/*
int _imageFilter::MIDfilter(const QVector<QRgb> &colorTable,QList<int> midval){
 for(int i=0;i<midval.size();i++)
   for(int j=0;j<midval.size()-1-i;j++)
     if(colorTable.at(midval.at(j))>colorTable.at(midval.at(j+1)))
       midval.swap(j,j+1);
 return midval.at(int(midval.size()/2));
}
*/

void _imageFilter::setDiv()
{
 _div=0;
 for(int i=0;i<_weight.size();i++)_div+=_weight.at(i);
 if(!_div)_div=1;
}


void _imageFilter::setWeight(Name name)
{
    _weight.clear();

    switch (name) {
   //Нет фильтра
    case Empty:{
        for(int i=-_edge;i<_edge+1;i++)
            for(int j=-_edge;j<_edge+1;j++)
                if(!i && !j)
                    _weight.append(1);//Эквивалент (i==0)&(j==0)
                else
                    _weight.append(0);
    };break;

    //Блюр
    case Blur:{
        for (int i = -_edge; i < _edge + 1; ++i)
            for (int j = -_edge; j < _edge + 1; ++j)
                if (i == -_edge || i == _edge || j == -_edge || j == _edge) {
                    if (i == 0 || j == 0)
                        _weight.append(1);
                    else
                        _weight.append(0);
                } else
                    _weight.append(1);
    };break;

    case MotionBlur:{
        for (int i = -_edge; i < _edge + 1; ++i)
            for (int j = -_edge; j < _edge + 1; ++j)
                if(i == j)
                    _weight.append(1);
                else
                    _weight.append(0);
    };break;

   //Тиснение (лёгкое)
    case Emboss:{
             for(int i=-_edge;i<_edge+1;i++)
               for(int j=-_edge;j<_edge+1;j++){
                 if(i==j){
                   if(i<0)
                       _weight.append(-1);
                   else
                       _weight.append(1);
                 } else
                     _weight.append(0);
               }
    };break;

   //---Резкость
    case Sharpen:{
        for (int i = -_edge; i < _edge + 1; ++i)
            for (int j = -_edge; j < _edge + 1; ++j)
                if(i == 0 && j == 0)
                    _weight.append(qPow((float)(2 * _edge + 1), 2) - 1);//_weight.append(4*_edge*mul+1);
                else {
                    if(!i || !j)
                        _weight.append(-1);
                    else
                        _weight.append(0);
                }
    };break;

   //Контуры
    case Contouring:{
        for (int i = -_edge; i < _edge + 1; ++i)
            for (int j = -_edge; j < _edge + 1; ++j)
                if(i == 0 && j == 0)
                    _weight.append(qPow((float)(2 * _edge + 1), 2) - 1);
                else {
                    if(!i || !j)
                        _weight.append(0);
                    else
                        _weight.append(-1);
                }
    };break;

    }

    //qDebug() << _weight;

    setDiv();
}



float _imageFilter::scaleFit(const QSize &size,int div){
//Эта ф-ция нужна для масштабирования, т.к. QImage не работает с большими картинками
//Ф-ция масштабирует большие картинки, размер которых слишком велик, к максимально возможному размеру
  float scale=1;

  qreal pixelCount=qreal(size.height())*qreal(size.width());
  if(div==0)div=128;
  float maxPixelCount=INT_MAX/div;//По идее INT_MAX/4 т.к. QImage::Format_ARGB32_Premultiplied занимает 32 бита, т.е. 4 байта
                                 //Эмпирическим путём взято 128)))
                                 //См. http://stackoverflow.com/questions/7080052/qimage-qpixmap-size-limitations

  if(pixelCount>=maxPixelCount){
    int maxwh=(size.height()>size.width())?size.height():size.width();
    scale = qPow((double)maxPixelCount, 0.5);
    scale/=maxwh;
  }

  return scale;
}

